@@ -115,10 +115,12 @@ class BlackboxTestCase(TestCase):
 
 def self_test_suite():
     names = [
+        "annotate",
         "archive",
         "blackbox",
         "bundle",
         "cli",
+        "cli_cherry_pick",
         "cli_merge",
         "client",
         "cloud_gcs",
@@ -126,8 +128,10 @@ def self_test_suite():
         "config",
         "credentials",
         "diff_tree",
+        "dumb",
         "fastexport",
         "file",
+        "gc",
         "grafts",
         "graph",
         "greenthreads",
@@ -140,19 +144,24 @@ def self_test_suite():
         "lru_cache",
         "mailmap",
         "merge",
+        "missing_obj_finder",
+        "notes",
         "objects",
         "objectspec",
         "object_store",
-        "missing_obj_finder",
         "pack",
         "patch",
         "porcelain",
+        "porcelain_cherry_pick",
         "porcelain_merge",
+        "porcelain_notes",
         "protocol",
+        "rebase",
         "reflog",
         "refs",
         "repository",
         "server",
+        "sparse_patterns",
         "stash",
         "submodule",
         "utils",
