@@ -500,11 +500,18 @@ class ApplyIncludedPathsTests(TestCase):
         self.assertTrue(idx[b"test_file.txt"].skip_worktree)
 
     def test_local_modifications_ioerror(self):
-        """Test handling of IOError when checking for local modifications."""
+        """Test handling of PermissionError/OSError when checking for local modifications."""
+        import sys
+
         self._commit_blob("special_file.txt", b"content")
         file_path = os.path.join(self.temp_dir, "special_file.txt")
 
-        # Make the file unreadable
+        # On Windows, chmod with 0 doesn't make files unreadable the same way
+        # Skip this test on Windows as the permission model is different
+        if sys.platform == "win32":
+            self.skipTest("File permissions work differently on Windows")
+
+        # Make the file unreadable on Unix-like systems
         os.chmod(file_path, 0)
 
         # Add a cleanup that checks if file exists first
@@ -517,8 +524,8 @@ class ApplyIncludedPathsTests(TestCase):
 
         self.addCleanup(safe_chmod_cleanup)
 
-        # Should raise conflict error with unreadable file and force=False
-        with self.assertRaises(SparseCheckoutConflictError):
+        # Should raise PermissionError with unreadable file and force=False
+        with self.assertRaises((PermissionError, OSError)):
             apply_included_paths(self.repo, included_paths=set(), force=False)
 
         # With force=True, should remove the file anyway
